//Element.h

///
/// This is the base class for all elements of the grammar
/// Any element that is a nonterminal uses this base class
/// Also sytactical elements such as commas can use this class
/// Elements that perform operations or return values should inherit
/// from this class (see Terminals.h)
///


#ifndef __ELEMENT_H__
#define __ELEMENT_H__

#include<string>
#include "Stringmanip.h"
#include<vector>
#include<math.h>
#include<deque>
#include "defines.h"

#include<iostream> // remove later

using namespace std;

class Element{
  
  public:
 
    /// constructor for nonterminal
    // Element(string symbol, bool terminal_status = false, int priority_level=0);		//S: removing priority_level
	Element(string symbol, bool terminal_status = false);
    /// constructor for terminal
    // Element(DATATYPE value, bool terminal_status, int priority_level=0);	//S: removing priority_level
	Element(DATATYPE value, bool terminal_status);
    virtual ~Element(){};
    
    /// Returns true for is a terminal
    bool get_isterminal() const;

    ///S: Returns true if an element is a variable
    bool get_isVariable() const;

    /// Returns number of arguments needed by this element		//S: not required
    // int get_num_args() const;

    /// Sets the number of arguments		//not required
    // void set_num_args(int nargs);

    /// Returns symbol showing which element this is
    string get_name() const;

    /// Returns priority level for evaluating this element		//S: not required
    // int get_priority() const;
    
    /// scale result using sigmoid function				//S: not required
    // static DATATYPE ActivateSigmoid(DATATYPE x);    
    /// adjust results for infinite or nan
    // static DATATYPE AdjustResult(DATATYPE x);
    
    /// nonterminals return 0 for evaluation
    virtual DATATYPE evaluate(){return 0;};	//S: evaluate() doesn't take any arguments

    friend ostream & operator << (ostream & os, const Element & el);
  
  protected:
     string name;
     bool isterminal, var_args;
     int num_args, priority;
	 
	 //variable-check
	 bool isVariable;
	 

};


#endif   
