// $Header: /usr5/cvsroot/nprojects/motsinger/genn/genn/src-mpi/GARandomSearchGA.cpp,v 1.4 2007/10/25 20:06:07 cvs Exp $
/* ----------------------------------------------------------------------------
  gasimple.C
  mbwall 28jul94
  Copyright (c) 1995 Massachusetts Institute of Technology
                     all rights reserved

  Source file for the simple genetic algorithm object.
---------------------------------------------------------------------------- */
#include "GARandomSearchGA.h"
#include <ga/garandom.h>
// included for this genetic algorithm
// needed in step function
#include "genome.h"

extern Translator trans;
extern Tester genn_test;
extern Grammar gram;


GAParameterList&
GARandomSearchGA::registerDefaultParameters(GAParameterList& p) {
  GAGeneticAlgorithm::registerDefaultParameters(p);

  p.add(gaNelitism, gaSNelitism,
	GAParameter::BOOLEAN, &gaDefElitism);

  return p;
}

GARandomSearchGA::GARandomSearchGA(const GAGenome& c) : GAGeneticAlgorithm(c){
  oldPop = pop->clone();

  el = gaTrue;
  params.add(gaNelitism, gaSNelitism, GAParameter::BOOLEAN, &el);
  debugOn = false;
}
GARandomSearchGA::GARandomSearchGA(const GAPopulation& p) : GAGeneticAlgorithm(p){
  oldPop = pop->clone();

  el = gaTrue;
  params.add(gaNelitism, gaSNelitism, GAParameter::BOOLEAN, &el);
  debugOn = false;
}
GARandomSearchGA::GARandomSearchGA(const GARandomSearchGA& ga) : GAGeneticAlgorithm(ga){
  oldPop = (GAPopulation *)0;
  copy(ga);
  debugOn = false;
}
GARandomSearchGA::~GARandomSearchGA(){
  delete oldPop;
  if(debugOn)
    debugstream.close();
}
GARandomSearchGA&
GARandomSearchGA::operator=(const GARandomSearchGA& ga){
  if(&ga != this) copy(ga); 
  return *this;
}
void 
GARandomSearchGA::copy(const GAGeneticAlgorithm & g){
  GAGeneticAlgorithm::copy(g);
  const GARandomSearchGA& ga = DYN_CAST(const GARandomSearchGA&,g);
  el = ga.el;
  debugOn = ga.debugOn;
  if(oldPop) oldPop->copy(*(ga.oldPop));
  else oldPop = ga.oldPop->clone();
  oldPop->geneticAlgorithm(*this);
}


int
GARandomSearchGA::setptr(const char* name, const void* value){
  int status = GAGeneticAlgorithm::setptr(name, value);

  if(strcmp(name, gaNelitism) == 0 ||
     strcmp(name, gaSNelitism) == 0){
    el = (*((int*)value) != 0 ? gaTrue : gaFalse);
    status = 0;
  }
  return status;
}

int
GARandomSearchGA::get(const char* name, void* value) const {
  int status = GAGeneticAlgorithm::get(name, value);

  if(strcmp(name, gaNelitism) == 0 || 
     strcmp(name, gaSNelitism) == 0){
    *((int*)value) = (el == gaTrue ? 1 : 0);
    status = 0;
  }
  return status;
}

void 
GARandomSearchGA::objectiveFunction(GAGenome::Evaluator f){
  GAGeneticAlgorithm::objectiveFunction(f);
  for(int i=0; i<pop->size(); i++)
    oldPop->individual(i).evaluator(f);
}

void 
GARandomSearchGA::objectiveData(const GAEvalData& v){
  GAGeneticAlgorithm::objectiveData(v);
  for(int i=0; i<pop->size(); i++)
    pop->individual(i).evalData(v);
}

const GAPopulation&
GARandomSearchGA::population(const GAPopulation& p) {
  if(p.size() < 1) {
    GAErr(GA_LOC, className(), "population", gaErrNoIndividuals);
    return *pop;
  }

  GAGeneticAlgorithm::population(p);
  GAPopulation * newpop = pop->clone();
//  oldPop->copy(*pop->clone());
  oldPop->copy(*newpop);
  oldPop->geneticAlgorithm(*this);
  delete newpop;
  
  return *pop;
}


///
/// population is always size of 1
/// and contains the best genome found
/// so far
///
int 
GARandomSearchGA::populationSize(unsigned int n) {
  effectivePopSize = n;
  n = 1;
  GAGeneticAlgorithm::populationSize(n);
  oldPop->size(n);
  return n;
}

int 
GARandomSearchGA::minimaxi(int m) { 
  GAGeneticAlgorithm::minimaxi(m);
  if(m == MINIMIZE)
    oldPop->order(GAPopulation::LOW_IS_BEST);
  else
    oldPop->order(GAPopulation::HIGH_IS_BEST);
  return minmax;
}


// Initialize the population, set the random seed as needed, do a few stupidity
// checks, reset the stats.  We must initialize the old pop because there is no
// guarantee that each individual will get initialized during the course of our
// operator++ operations.  We do not evaluate the old pop because that will 
// happen as-needed later on.
void
GARandomSearchGA::initialize(unsigned int seed)
{
  GARandomSeed(seed);

  pop->initialize();
  pop->evaluate(gaTrue);	// the old pop will get it when the pops switch
//  oldPop->initialize();

  stats.reset(*pop);

  if(!scross) 
    GAErr(GA_LOC, className(), "initialize", gaErrNoSexualMating);
}


//   Evolve a new generation of genomes.  When we start this routine, pop
// contains the current generation.  When we finish, pop contains the new 
// generation and oldPop contains the (no longer) current generation.  The 
// previous old generation is lost.  We don't deallocate any memory, we just
// reset the contents of the genomes.
//   The selection routine must return a pointer to a genome from the old
// population.

// Every step merely generates a new population
// Best individual across all generations is saved
void
GARandomSearchGA::step()
{

	cerr << "DEBUG:  GARandomSearchGA::step() called\n";
// keep one single best genome in pop
// create random codon lists and compare
// at end compare best of these with the best genome
// if score is better than best genome
// replace single genome with the new best
  bool fullTree = true;
  vector<unsigned int> codons;
  int currInd=1;
  
  float bestscore = pop->best().score();
  vector<unsigned int> bestCodons;
  
  // output best as first in each generation
  vector<unsigned int> tempCodons;
  
  if(debugOn){
    debugstream <<  "STEP" << endl;
    outputIndividual(pop->best(), 0);
  }
  
//GEBinStrBlockCrossGenome newGenome(1, GenomeEvaluator);

  // need to create individuals up to effectivePopSize
  for(int currGenome=0; currGenome<effectivePopSize; currGenome++){
    float score = randomGenomeCreate(codons, fullTree);

    fullTree = !fullTree;
    if(score > bestscore){
      bestscore = score;
      bestCodons = codons;
    }
    
    if(debugOn){
      outputIndividual(codons, currInd++, score);
    }
  }
  

  // when best score is better than the score in 
  // the population we need to replace that genome
  if(bestscore > pop->best().score()){
    GEBinStrBlockCrossGenome &bestGenome=DYN_CAST(GEBinStrBlockCrossGenome &, pop->best());
    // need to resize the best
    bestGenome.resize(bestCodons.size() * bestGenome.CODONSIZE);
    // now use masks to test bits in the codons vector and
    // set the bits to match
    // start at highest bit position
    vector<unsigned int> masks;
    for(int i=bestGenome.CODONSIZE-1; i>=0; i--){
      masks.push_back((unsigned int)pow(float(2), i));
    }    
    vector<unsigned int>::iterator maskIter;
    unsigned int geneCounter=0;
    unsigned int lastCodon = bestCodons.size();
    for(unsigned int currCodon=0; currCodon < lastCodon; currCodon++){
      for(maskIter=masks.begin(); maskIter != masks.end(); maskIter++){
        bestGenome.gene(geneCounter++, *maskIter & bestCodons[currCodon]);
      } 
    }
    bestGenome.score(bestscore);
  }

  
}

// returns the fitness value for this set of codons
// codons are stored in the input variable codons
float GARandomSearchGA::randomGenomeCreate(vector<unsigned int> & codons,
  bool fullTree){
  codons.clear();
  
  int maxDepth = gram.get_init_tree_depth();
  // create tree that initializes genome
  // tree will be full half the time and grown half the time                 
  // as described by koza
  fullTree = !fullTree;
  tree_node start;
  
  start.nodeElement = gram.get_start();
  start.ruleIndex = get_rule_index(start.nodeElement, maxDepth, fullTree);            

  tree<tree_node> tr(start);        
  tree<tree_node>::iterator top;
  top=tr.begin(); 

  add_rule_nodes(tr, top, maxDepth, fullTree);
  // after tree is constructed need to build the linear version
  // of the tree as the initial genome               
  tree<tree_node>::iterator currNode;
  for(currNode = tr.begin(); currNode != tr.end(); currNode++){
    // generate value within codoonsize that will give modulus desired
    codons.push_back((*currNode).ruleIndex);
  }
  // check to see if tree exceeds max size of genome
  // when tree exceeds number of codons, fill up as far as possible within the 
  // maximum size
  GEBinStrBlockCrossGenome &best=DYN_CAST(GEBinStrBlockCrossGenome &, pop->best());
  if(tr.size() * best.CODONSIZE > best.maxX){

    int lastCodon = best.maxX / best.CODONSIZE;
    // truncate codons to fit genome size
    codons.erase(codons.begin()+lastCodon, codons.end());
    return fitnessEvaluator(codons);
    // changed so that the genome will be truncated
    // as it would be in standard GENN and the 
    // fitness will still be evaluated
    //return 0.0; // set this fitness to zero as it is too long
  }
  else{
    // check fitness
    return fitnessEvaluator(codons);    
  }
}

// evaluates codons
float GARandomSearchGA::fitnessEvaluator(std::vector<unsigned int> & codons){
  float total=0.0;
  // convert genome into grammar rules
  trans.set_codons(codons);
  vector<Element *> tree;
  // vector<Element *> postfix_stack;		//S: Not required
  vector<Element *> constants_list; // used to track dynamically created constants
  if(trans.create_terminals(tree, constants_list)){
	  
    // convert to postfix stack
    // postfix_stack=trans.create_postfix_stack(term_list);		//S: Not required
	
    // need to test the genome
    total = 1 - genn_test.determine_error_rate(tree);
  }  

  for(unsigned int conCounter=0; conCounter<constants_list.size(); conCounter++){
      delete constants_list[conCounter];
  }

  return total;
}


// outputs individual for debugging
void GARandomSearchGA::outputIndividual(vector<unsigned int> & codons, int currInd, float fitness){
  debugstream << "Individual #" << currInd << endl;

  trans.set_codons(codons);
  for(unsigned int i=0; i<codons.size(); i++)
    debugstream << codons[i] << " ";
    
  debugstream << endl;
    
  vector<Element *> term_list;
  vector<Element *> constant_list;
  int effectiveSize = trans.create_terminals(term_list, constant_list);
  
  
  if(effectiveSize > 0){
    for(unsigned int i=0; i<term_list.size(); i++){
      if((term_list[i]->get_name())[0] == 'V'){
        debugstream << term_list[i]->get_name() << " ";
      }
    }  
    debugstream << endl;
  }

  debugstream << "genome Size=" << codons.size() << "\teffective Size=" << effectiveSize << endl;
  debugstream << "fitness=" << fitness << endl;
  // delete any constants created earlier
  for(unsigned int conCounter=0; conCounter<constant_list.size(); conCounter++){
    delete constant_list[conCounter];
  }
}




// For debugging outputs all information for all members 
// of the population
// Arg:  population to output
void GARandomSearchGA::outputIndividual(GAGenome & g, int currInd){
  
  GEBinStrBlockCrossGenome &genome=DYN_CAST(GEBinStrBlockCrossGenome &, g);
  // need to output codons, score, and variables present
  debugstream << "Individual #" << currInd << endl;

  trans.convert_codons(genome);

  vector<unsigned int> codons = trans.get_codon_values();
  for(unsigned int i=0; i<codons.size(); i++)
    debugstream << codons[i] << " ";
  debugstream << endl;
    
  vector<Element *> term_list;
  vector<Element *> constant_list;
  int effectiveSize = trans.create_terminals(term_list, constant_list);
    
  for(unsigned int i=0; i<term_list.size(); i++){
    if((term_list[i]->get_name())[0] == 'V'){
      debugstream << term_list[i]->get_name() << " ";
    }
  }

  debugstream << endl;
  debugstream << "genome Size=" << codons.size() << "\teffective Size=" << effectiveSize << endl;
  debugstream << "fitness=" << genome.score() << endl;
  // delete any constants created earlier
  for(unsigned int conCounter=0; conCounter<constant_list.size(); conCounter++){
    delete constant_list[conCounter];
  }
}
