// InputReader.cpp

#include "InputReader.h"
#include "mpi.h"
#include "unistd.h"

///
/// Reads input data file, records status of individuals in
/// a Tester, and records all variable values in Variable
/// @param inputfile Data file with variables and individual
/// status
/// @param testobj Tester for storing status and testing
/// neural networks that are created
///
void InputReader::read_input(string inputfile, Tester & testobj){

  /*cerr << "Sleep debugging\n";
    int myrank = -1;
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
    sleep(myrank  * 30);
    cerr << "Awaken, fair princess (" << myrank << ")...\n";*/
  
  cerr << "About to open input file\n";
  ifstream input(inputfile.c_str(), ios::in);
  if(!input.is_open()){
    throw ErrorExcept(inputfile + ":  cannot open!");
  }
  cerr << "Opened input file\n";
   
  char line[LINEMAX];
  int linecounter = 0;
  string current_line;
  Variable::clear_variable_list();
  testobj.clear_status();
  
  while(!input.eof()){
    linecounter++;
    input.getline(line, LINEMAX);
    
    string current_line(line);
    int linelength;
    if(!(linelength = current_line.length())) {
      continue; // blank line so go to next one in file    
		}
    vector<string> tokens;
    //split the string on either space or tab
    // S: first (S: not last) token will be the status of the individual (either 0 or 1)
    if(Stringmanip::split_string(current_line, " ", "\t", tokens) > 0){
      if(!(Stringmanip::isnumber(tokens.back()))) {
        cerr << "On line " << linecounter << ":\n";
        throw ErrorExcept("ERROR: " + tokens.back() + " is not a number in " + inputfile);
      }
	    
      //Slightly wasteful, but this memory is cheap...
	  
      // testobj.add_status_quan(Stringmanip::stodata(tokens.back()));		//S:
      // testobj.add_status(Stringmanip::stoi(tokens.back()));  
		testobj.add_status(Stringmanip::stoi(tokens.front()));
    }
    
    vector<INPUT_TYPE> data;
	//S: Variables start from 2nd token to the last token
	
    //S: for(unsigned int i=0; i<tokens.size()-1; i++){
	for(unsigned int i=1; i<=tokens.size()-1; i++){
      data.push_back(Stringmanip::stoinput(tokens[i]));

			INPUT_TYPE tmp = Stringmanip::stoinput(tokens[i]);

			
			if ( (-1 != tmp) && (0 != tmp) && (1 != tmp) && (2 != tmp) &&
					 (3 != tmp) ) {
				cerr << "ERROR:  unexpected data value of " << static_cast<INPUT_TYPE>(tmp) << " (" <<
					tokens[i] << ") encountered!\n";
				exit(1);
			}

			/*
			cerr << "ERROR:  expected data value of " << tmp << " (" <<
					tokens[i] << ") encountered!\n";
				exit(1);
			*/

			//Variable::var_values.push_back(Stringmanip::stodata(tokens[i]));
    }
    
    Variable::fill_variable_list(data);    
  }
  input.close();
  
  cerr << "Finished reading input\n";

}
