/* ----------------------------------------------------------------------------
  DemeGA.C
  mbwall 28jul94
  Copyright (c) 1995-1996 Massachusetts Institute of Technology
                          all rights reserved

   Souce file for the deme-based genetic algorithm object.
---------------------------------------------------------------------------- */
#include <string.h>
#include <ga/garandom.h>
#include "PVMDemeGA.h"


//#define DEBUG

PVMDemeGA::PVMDemeGA(const GAGenome& c) : GADemeGA(c) {
  _status = 0;
  _Ntid = _ntid = _nhosts = _narch = _currgen =  0;
  _mid = 0;

}
PVMDemeGA::PVMDemeGA(const GAPopulation& p) : GADemeGA(p) {
  _status = 0;
  _Ntid = _ntid = _nhosts = _narch = _currgen = 0;
  _mid = 0;
}
PVMDemeGA::PVMDemeGA(const PVMDemeGA& orig) : GADemeGA(orig) {
  _status = 0;
  _Ntid = _ntid = _nhosts = _narch = _currgen = 0;
  _mid = 0;
  copy(orig);
}
PVMDemeGA::~PVMDemeGA(){

}
PVMDemeGA&
PVMDemeGA::operator=(const PVMDemeGA& orig){
  if(&orig != this) copy(orig); 
  return *this;
}
void
PVMDemeGA::copy(const GAGeneticAlgorithm& g){
  GADemeGA::copy(g);
}

// determines and passes seed for each
// process that was spawned
// seed is offset from original seed of 
// config file unless it is greater than INT_MAX
// in which case it starts as offset from zero
void PVMDemeGA::set_seeds(unsigned int seed){
	//now set random seeds for each slave
  //and wait for responses from slaves
  //that message received
  for(int j=1; j<=_ntid; j++) {
    int eachseed;
    if(((unsigned int)(INT_MAX - seed)) >= (unsigned int)j){
      eachseed = j + seed;
    }
    else{
      eachseed = j - (INT_MAX - seed);
    }
		MPI_Send(&eachseed, 1, MPI_INT, j, MSG_SET_RANDOM_SEED, MPI_COMM_WORLD);
  }
  
  
  #ifdef DEBUG
  cerr << "waiting for response from slaves...\n";
#endif
  
}

// To initialize this genetic algorithm, we make sure all of our spawns are
// not evolving then we tell them all to initialize themselves.  Then we wait
// and copy their populations into our own.  Update our stats based upon the
// harvested populations.
void
PVMDemeGA::initialize(unsigned int seed) {

	/*dead...*/

}

// sets the interval at which migration occurs
void PVMDemeGA::set_pvm_exchange_gen(int num_gens){
  _pvm_exchange_gen = num_gens;
}


// To evolve the genetic algorithm, we loop through all of our populations and
// tell each process to evolve its population for a certain number of 
// generations.  Then allow the migrator to do its thing.  Each process is
// supposed to keep track of the statistics for its population, so we reap 
// those as well.
void
PVMDemeGA::step() {

#ifdef DEBUG
  cerr << "sending step command to slaves...\n";
#endif

  for(int j=1; j<=_ntid; j++) {
    int nsteps = _pvm_exchange_gen;
		_status = MPI_Send(&nsteps, 1, MPI_INT, j, MSG_STEP, MPI_COMM_WORLD);
		if (MPI_SUCCESS != _status) {
			cerr << "DEBUG:  error in PVMDemeGA::step () MPI_Send to " << j
					 << " : " << _status << "\n";
		}
  } /*End telling slaves to step*/
	

	/*abstract the steps in gennMPI::gennSlave into a seperate function,
		then have master run through a step here?  Remove slave from loop above...
	  can MPI processes talk to themselves?  Would make for a better abstraction,
	  as long as it doesn't cause deadlocks...*/

#ifdef DEBUG
  cerr << "waiting for slaves to step...\n";
#endif
 
  int flag = _ntid;
	int ival = 0;
	MPI_Status Stat;

  while(flag > 0) {
		int rc = MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &Stat);
    if(MPI_SUCCESS == rc) {
      int msgtag, tid;
			msgtag = Stat.MPI_TAG;
			tid = Stat.MPI_SOURCE;

      switch(msgtag) {
      case MSG_STEP_COMPLETE:
				MPI_Recv(&ival, 1, MPI_INT, tid, MSG_STEP_COMPLETE,
								 MPI_COMM_WORLD, &Stat);
				flag--;

#ifdef DEBUG
	cerr << "  tid " << tid << " has finished step\n";
#endif
	break;

			case MSG_SLAVE_EXITED:
				cerr << "DEBUG:  MSG_SLAVE_EXITED - this should never happen\n";
				exit(3);
				
        break;
      default:
				cerr << className() << ": step:\n";
				cerr << "  unexpected msgtag7: " << msgtag << " from " << tid << "\n";
				cerr << "  PVMDemeGA exiting\n";
				exit(1);
				break;
      }
    }
    else {
      cerr << className() << ": step:\n";
      cerr << "  error from MPI_Probe: " << rc << "\n";
    }
  }

  if(++_currgen < this->nGenerations()){
    migrate();
  }

}  /*End PVMDemeGA::step()*/

// smd -- added to override base class done 
// now no need to collect deme populations and
// run stats on them
int PVMDemeGA::done(){
  if(_currgen < this->nGenerations()){
    return 0;
  }
  return 1;
}

// Altered to send the single best individual from each population
// to every other population (process)

// This is the island model

// The migrator tells the spawned populations when, where, and how many 
// individuals to move around.  This example does a single migration from one
// randomly chosen population to another randomly chosen population.  It
// chooses randomly the number of individuals to migrate.
//   One simple way to get your own migration algorithm into this is to derive
// a genetic algorithm from this class then define your own migrate method.


// change so that master collects all individuals and then sends them
// to each node
void
PVMDemeGA::migrate() {

  collectMigration();

  // now send best genomes to each slave
  // don't send genome received from slave
  int numGenomesMigrate = _ntid;
  int count = _ntid-1;
  for(int i=0; i<numGenomesMigrate; i++) {
    // pack and send all inds except one received
    // from this slave
		int toid = i + 1;  //inconsistent...
		
		int numbits = count;
		int * intbuf = 0;
		int bufindex = 0;
		float * floatbuf = (float*)malloc(count * sizeof(float));
		int sizes[2];
		int rc = 0;
		if (0 == floatbuf) {
			cerr << "ERROR:  malloc returned 0 for floatbuf in PVMDemeGA::migrate\n";
			exit(1);
		}
		
		/*Determine how many bits are in the individuals we're sending*/
		int pindex = 0;
		for (int j = 0; j < numGenomesMigrate; j++) {
			if (j != i) {
				numbits += ((GEBinStrBlockCrossGenome&)(migratingGenomes[j])).length();
				floatbuf[pindex++] = ((GEBinStrBlockCrossGenome&)(migratingGenomes[j])).score();
			}
		}
		
		/*Pack the individuals*/
		intbuf = (int*)malloc(numbits * sizeof(int));
		if (0 == intbuf) {
			cerr << "ERROR:  malloc returned 0 for intbuf in PVMDemeGA::migrate\n";
			exit(1);
		}
		pindex = 0;
		for (int j = 0; j < numGenomesMigrate; j++) {
			if (j != i) {
				bufindex = PackIndividualBits(migratingGenomes[j], intbuf, bufindex);
			}																				
		}
		
		sizes[0] = count;
		sizes[1] = numbits;
		
		/*Send to the slave*/
		rc = MPI_Send(sizes, 2, MPI_INT, toid, MSG_INCOMING_MIGRATION,
									MPI_COMM_WORLD);
		
		rc = MPI_Send(intbuf, numbits, MPI_INT, toid,
												MSG_INCOMING_MIGRATION_BITS,
												MPI_COMM_WORLD);
		
		rc = MPI_Send(floatbuf, count, MPI_FLOAT, toid,
												MSG_INCOMING_MIGRATION_SCORES,
												MPI_COMM_WORLD);

		free(intbuf);
		free(floatbuf);

	} //end foreach pop


	/*Print statistics of best individual received*/
	float bestscore = 0;
	int bestindex = 0;
	for (int i = 0; i < numGenomesMigrate; i++) {
		if (((GEBinStrBlockCrossGenome&)(migratingGenomes[i])).score() > bestscore) {
			bestscore = ((GEBinStrBlockCrossGenome&)(migratingGenomes[i])).score();
			bestindex = i;
		}
	}
	
	cerr << "BESTFITNESS: " << bestscore << "\n";

} //end migrate 


// Gather the population data from each of our distributed populations.
int
PVMDemeGA::collect() {

  cerr << "DEBUG:  PVMDemeGA::collect() called - check msg_send_best\n";
#ifdef DEBUG
  cerr << "sending request for populations...\n";
#endif

	int nothing;
for(int j=1; j<=_ntid; j++) {
	
	_status = MPI_Send(&nothing, 1, MPI_INT, j, MSG_SEND_BEST,
										 MPI_COMM_WORLD);
 }
 
#ifdef DEBUG
  cerr << "waiting for populations from slaves...\n";
#endif

  int flag = _ntid;
	MPI_Status Stat;
  while(flag > 0) {
		int rc = MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD,
											 &Stat);
    //if(bufid >= 0) {
		if (MPI_SUCCESS == rc) {
      int msgtag, tid;
			msgtag = Stat.MPI_TAG;
			tid = Stat.MPI_SOURCE;
			int which = tid - 1;

      switch(msgtag) {
      case MSG_INCOMING_POPULATION:
				_status = RecvPopulation(*deme[which], tid);
				flag--;

#ifdef DEBUG
				cerr << "  received pop from tid " << tid << " (" << which << ")\n";
#endif
				break;

      default:
				cerr << className() << ": collect:\n";
				cerr << "  unexpected msgtag9: " << msgtag << "\n";
				break;
      } /*End switch(msgtag)*/
			} /*end if MPI_SUCCESS*/
			else {
				cerr << className() << ": collect:\n";
				cerr << "  error from MPI_Recv: " << rc << "\n";
			}
  } //end while flag > 0
	
  return _status;
} /*End collect*/


// Gather the best individuals from each population
// and return the best one in the input variable
int PVMDemeGA::collectMigration(){
#ifdef DEBUG
  cerr << "sending request for best individuals...\n";
#endif 

	int nothing = 0;
	
   for(int j=1; j<=_ntid; j++) {
		 _status = MPI_Send(&nothing, 1, MPI_INT, j, MSG_SEND_BEST,
												MPI_COMM_WORLD);
  }

#ifdef DEBUG
  cerr << "waiting for individuals from slaves (collectMigration)...\n";
#endif
  //int individualCount; /*only one individual coming in*/
  GEBinStrBlockCrossGenome currentGenome(1, GenomeEvaluator);
  int flag = _ntid;
	MPI_Status Stat;
  while(flag > 0) {
		int rc = MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &Stat);
    if(MPI_SUCCESS == rc) {
      int msgtag, tid;
			int sizes[2];
			int * intbuf;
			float inscore = 0.0;
			MPI_Status Stat2;
			int rc2 = 0;

			msgtag = Stat.MPI_TAG;
			tid = Stat.MPI_SOURCE;
			int which = tid - 1;

      //int which = tid2idx(tid);
      switch(msgtag) {
      case MSG_INCOMING_MIGRATION:
				
				rc2 = MPI_Recv(&sizes, 2, MPI_INT, tid, MSG_INCOMING_MIGRATION,
													 MPI_COMM_WORLD, &Stat2);

				intbuf = (int*)malloc(sizes[1] * sizeof(int));

				if (0 == intbuf) {
					cerr << "ERROR: memory error in PVMDemeGA::collectMigration()\n";
					exit(1);
				}

				rc2 = rc2 + MPI_Recv(intbuf, sizes[1], MPI_INT, tid,
															MSG_INCOMING_MIGRATION_BITS,
															MPI_COMM_WORLD, &Stat2);

				rc2 = rc2 + MPI_Recv(&inscore, 1, MPI_FLOAT, tid,
															MSG_INCOMING_MIGRATION_SCORES,
															MPI_COMM_WORLD, &Stat2);

				if (MPI_SUCCESS != rc) {
					cerr << "ERROR:  Sum of incoming migration rc2: " << rc2 << "\n";
				}

				UnpackIndividualBits(migratingGenomes[which], intbuf, 0);
				migratingGenomes[which].score(inscore);

				free(intbuf);

	      flag--;

#ifdef DEBUG
	cerr << "  received individual from tid " << tid << " (" << which << ")\n";
#endif
	    break;

      default:
	      cerr << className() << ": collect:\n";
	      cerr << "  unexpected msgtag1: " << msgtag << "\n";
	      break;
      }
    } //end if successful MPI_Probe
    else {
      cerr << className() << ": collectBest:\n";
      cerr << "  error from MPI_Recv: " << rc << "\n";
    }
  }  //End while flag > 0

  return _status;
}


// Gather the best individuals from each population
// and return the best one in the input variable
int PVMDemeGA::collectBest(GAGenome* bestGenome){
#ifdef DEBUG
  cerr << "sending request for best individuals...\n";
#endif  
   for(int j=1; j<=_ntid; j++) {
		 int nothing = 0;
		 _status = MPI_Send(&nothing, 1, MPI_INT, j, MSG_SEND_BEST, 
												MPI_COMM_WORLD);
  }

#ifdef DEBUG
  cerr << "waiting for individuals from slaves (collectBest)...\n";
#endif

	int individualCount = 1;
  GEBinStrBlockCrossGenome currentGenome(1, GenomeEvaluator);
  bool firstOne = true;
  int flag = _ntid;
	MPI_Status Stat;

  while(flag > 0) {
		int rc = MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &Stat);
    if(rc >= 0) {
      int msgtag, tid;
			int sizes[2];
			int * intbuf = 0;
			int bufindex = 0;
			float * floatbuf = 0;

			msgtag = Stat.MPI_TAG;
			tid = Stat.MPI_SOURCE;

      switch(msgtag) {
      case MSG_INCOMING_MIGRATION:
				
				_status = MPI_Recv(&sizes, 2, MPI_INT, tid, MSG_INCOMING_MIGRATION,
													MPI_COMM_WORLD, &Stat);
				
				intbuf = (int*)malloc(sizes[1] * sizeof(int));
				floatbuf = (float*)malloc(sizes[0] * sizeof(float));
				
				if ( (0 == intbuf) || (0 == floatbuf)) {
					printf("ERROR: memory error in genome.cpp::RecvMigration\n");
					exit(1);
				}
				
				_status = MPI_Recv(intbuf, sizes[1], MPI_INT, tid,
																		MSG_INCOMING_MIGRATION_BITS,
																		MPI_COMM_WORLD, &Stat);
				
				_status = MPI_Recv(floatbuf, sizes[0], MPI_FLOAT, tid,
																		MSG_INCOMING_MIGRATION_SCORES,
																		MPI_COMM_WORLD, &Stat);
				
				for (int i = 0; i < individualCount; i++) {
					bufindex = UnpackIndividualBits(currentGenome, intbuf, bufindex);
					((GEBinStrBlockCrossGenome&)(currentGenome)).score(floatbuf[i]);
				}

				free(intbuf);
				free(floatbuf);
					
				
				/*We just want the overall best one*/
				if(firstOne){
          firstOne = false;
          *bestGenome = currentGenome;
					cerr << "First score: " << currentGenome.score() << "\n";
        }
        else if(currentGenome.score() > bestGenome->score()){
          *bestGenome = currentGenome;
					cerr << "Best score: " << currentGenome.score() << "\n";
        }

	      flag--;

	    break;

      default:
	      cerr << className() << ": collect:\n";
	      cerr << "  unexpected msgtag2: " << msgtag << "\n";
				exit(2);
	      break;
      }
    }
    else {
      cerr << className() << ": collectBest:\n";
      cerr << "  error from mpi_probe: " << rc << "\n";
			exit(4);
    }
  }  

  return _status;
} /*End collectBest*/

/*Set the number of populations we have:
  In MPI, assume we get all that we ask for*/
void PVMDemeGA::setPopSize(int size) {
	_ntid = _Ntid = size;

	migratingGenomes = std::vector<GEBinStrBlockCrossGenome> (_Ntid,GEBinStrBlockCrossGenome(1, GenomeEvaluator));

}
