#include "Terminals.h"

///
/// Terminal creator class
///
Terminal_creator::Terminal_creator(){
  // TerminalMap["+"] = Add;		//S: not required
  // TerminalMap["-"] = Sub;
  // TerminalMap["*"] = Mult;
  // TerminalMap["/"] = Div;
  // TerminalMap["PA"] = padd;
  // TerminalMap["PS"] = psub;
  // TerminalMap["PM"] = pmult;
  // TerminalMap["PD"] = pdiv;
  // TerminalMap["W"] = wt;
  // TerminalMap["."] = dot;
  // TerminalMap["V#"] = var;
  // TerminalMap["Concat"] = concat;
  // TerminalMap["PAND"] = pand;
  // TerminalMap["PNAND"] = pnand;
  // TerminalMap["POR"] = por;
  // TerminalMap["PNOR"] = pnor;
  // TerminalMap["PXOR"] = pxor;
  // TerminalMap["AND"] = booland;
  // TerminalMap["NAND"] = boolnand;
  // TerminalMap["OR"] = boolor;
  // TerminalMap["NOR"] = boolnor;
  // TerminalMap["XOR"] = boolxor;
	TerminalMap["V#"] = var;
	TerminalMap["0"] = val0;
	TerminalMap["1"] = val1;
	TerminalMap["2"] = val2;
	TerminalMap["+"] = pos;
	TerminalMap["-"] = neg;
}

///
/// Function for returning appropriate pointer based
/// on symbol passed in
/// @param symbol for selecting appropriate terminal
/// @param num_args number of arguments the terminal accepts
/// @return new Terminal 
///
Element * Terminal_creator::create_terminal(const string& symbol, int num_args){
    
    // check for variable 
    if(symbol[0] == 'V'){
      string num = symbol.substr(1,symbol.length()-1);
      if(Stringmanip::isnumber(num)){
        stringstream ss(num);
        int var_index;
        ss >> var_index;
        return create_variable(symbol, var_index);
      }
    }
    
    switch(TerminalMap[symbol]){
		case val0:
			return new Value0(symbol);
		case val1:
			return new Value1(symbol);
		case val2:
		  	return new Value2(symbol);
		case pos:
			return new Positive(symbol);
		case neg:
			return new Negative(symbol);
	  // case Add:
        // return new Addition(symbol, 2);		//S: not required
      // case Sub:
        // return new Subtraction(symbol, 2);
      // case Mult:
        // return new Multiplication(symbol, 2);
      // case Div:
        // return new Division(symbol, 2);
      // case padd:
        // return new pAdd(symbol, -1);
      // case psub:
        // return new pSub(symbol, -1);
      // case pmult: 
        // return new pMult(symbol, -1);
      // case pdiv:
        // return new pDiv(symbol, -1);
      // case wt:
        // return new Weight(symbol, 2);
      // case dot:
        // return new Dot(symbol, 0);
      // case concat:
        // return new ConCat(symbol, -1);
      // case pand:
        // return new pAnd(symbol, -1);
      // case pnand:
        // return new pNand(symbol, -1);
      // case por:
        // return new pOr(symbol, -1);
      // case pnor:
        // return new pNor(symbol, -1); 
      // case pxor:
        // return new pXor(symbol, -1);
      // case booland:
        // return new And(symbol, 2);
      // case boolnand:
        // return new Nand(symbol, 2);
      // case boolor:
        // return new Or(symbol, 2);
      // case boolnor:
        // return new Nor(symbol, 2); 
      // case boolxor:
        // return new Xor(symbol, 2);
      default:
        throw ErrorExcept(symbol + " is not defined as a valid symbol\nCheck Terminals.cpp\n\n");
    }
}

/// 
/// Creates a constant element
/// @param symbol  
///
Element * Terminal_creator::create_constant(const string& symbol){
  return new Constant(symbol,0);
}

///
/// Creates a variable
/// @param symbol for new Varible element
/// @param var_index index for variable in data vector
/// @return new Variable Element pointer
///
Element * Terminal_creator::create_variable(const string& symbol, int var_index){
  return new Variable(symbol, var_index);
}

//////////////////////////////////////
// Constant class
//////////////////////////////////////
Constant::Constant(string symbol, int number_args)
  :Element(symbol, true){
    num_args = number_args;
    // must convert symbol into value
    stringstream ss(symbol);
    ss >> constant_value;
}

Constant::Constant(DATATYPE value)
  :Element(value, true){
  num_args = 0;
  constant_value = value;
}

DATATYPE Constant::evaluate(){ 		//S: evaluate doesn't take any arguments
  return constant_value;
} 

//////////////////////////////////////
// Variable class
//////////////////////////////////////

vector<vector<INPUT_TYPE> > Variable::var_values;
int Variable::ind_index=0;
int Variable::missing_value = 3;

Variable::Variable(string symbol, int var_index)
  :Element(symbol, true){
    num_args = 0;
    // must convert symbol into value
    index_value = var_index-1;
	isVariable = true;		//S: variable-check
}

///
/// Returns value for the individual at the indicated location
/// @param args deque not used in this evaluation
/// @return value of the variable
/// @throws ErrorExcept when missing variable data at the indicated
/// locus for the indicated individual
///
  
//S: none of the evaluate methods require 'args' as parameter.. so removing it for all of them
// DATATYPE Variable::evaluate(deque<DATATYPE> & args){ 
 DATATYPE Variable::evaluate(){
  if ((Variable::var_values[Variable::ind_index][index_value]) == missing_value){
    throw ErrorExcept("Missing Data", ErrorExcept::missingData);
  }
  return Variable::var_values[Variable::ind_index][index_value];
} 

void Variable::fill_variable_list(vector<INPUT_TYPE> variables){
  Variable::var_values.push_back(variables);
}

void Variable::clear_variable_list(){
  Variable::var_values.clear();
}

void Variable::set_ind_index(int new_index){
  Variable::ind_index = new_index;
}

void Variable::set_missing_value(int m_value){
  Variable::missing_value = m_value;
}

///////////////////////////
//Value Classes//
////////////////////////////
	
Value0::Value0(string symbol)
	:Element(symbol,true){
		isVariable = false;		//S: added variable-check
	}

DATATYPE Value0::evaluate(){
	DATATYPE result = 0;		//S: hard-coded value for SNP = 0
	
	return result;
}

Value1::Value1(string symbol)
	:Element(symbol,true){
		isVariable = false;		//S: added variable-check
	}

DATATYPE Value1::evaluate(){
	DATATYPE result = 1;		//S: hard-coded value for SNP = 1
	
	return result;
}

Value2::Value2(string symbol)
	:Element(symbol,true){
		isVariable = false;		//S: added variable-check
	}

DATATYPE Value2::evaluate(){
	DATATYPE result = 2;		//S: hard-coded value for SNP = 2
	
	return result;
}

//////////////////////////////////////////
/*Classes for Class-statuses (+/-)*/
/////////////////////////////////////////

Positive::Positive(string symbol)
	:Element(symbol,true){
		isVariable = false;		//S: added variable-check
	}
	
DATATYPE Positive::evaluate(){
	DATATYPE result = 0;	//S: 0 = positive +
		
	return result;
}

Negative::Negative(string symbol)
	:Element(symbol,true){
		isVariable = false;		//S: added variable-check
	}
	
DATATYPE Negative::evaluate(){
	DATATYPE result = 1;	//S: 1 = negative -
		
	return result;
}


//S: not required
//////////////////////////////////////
// Addition class
//////////////////////////////////////

// Addition::Addition(string symbol, int number_args)
  // :Element(symbol, true, 2){
    // num_args = number_args;
// }
    
// DATATYPE Addition::evaluate(deque<DATATYPE> & args){
  // DATATYPE result = args[0];
  
  // for(int i=1; i<num_args; i++){
    // result += args[i];
  // }
  // return AdjustResult(result);
// } 

//////////////////////////////////////
// Subtraction class
//////////////////////////////////////

// Subtraction::Subtraction(string symbol, int number_args)
  // :Element(symbol, true, 2){
    // num_args = number_args;
// }
    
// DATATYPE Subtraction::evaluate(deque<DATATYPE> & args){
  // DATATYPE result = args[0];
  
  // for(int i=1; i<num_args; i++){
    // result -= args[i];
  // }
  // return AdjustResult(result);
// } 

//////////////////////////////////////
// Multiplication class
//////////////////////////////////////

// Multiplication::Multiplication(string symbol, int number_args)
  // :Element(symbol, true, 3){
    // num_args = number_args;
// }
    
// DATATYPE Multiplication::evaluate(deque<DATATYPE> & args){
  // DATATYPE result = args[0];
  
  // for(int i=1; i<num_args; i++){
    // if(args[i] == 0){
      // result = 0;
      // break;
    // } 
    // result *= args[i];
  // }
  // return AdjustResult(result);
// } 

//////////////////////////////////////
// Division class
//////////////////////////////////////
// Division::Division(string symbol, int number_args)
  // :Element(symbol, true, 3){
    // num_args = number_args;
// }
    
// DATATYPE Division::evaluate(deque<DATATYPE> & args){
  // DATATYPE result = args[0];
  
  // for(int i=1; i<num_args; i++){
    // if(args[i] == 0)
      // return 1.0;
    // result /= args[i];
  // }
  
  // return AdjustResult(result);
// } 

//////////////////////////////////////
// pAdd class
//////////////////////////////////////
// pAdd::pAdd(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE pAdd::evaluate(deque<DATATYPE> & args){
  // DATATYPE result = args[0];
  // for(unsigned int i=1; i<args.size(); i++){
    // result += args[i];
  // }
  
  // result = AdjustResult(result);
  
  // return ActivateSigmoid(result);
// } 

//////////////////////////////////////
// pSub class
//////////////////////////////////////
// pSub::pSub(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE pSub::evaluate(deque<DATATYPE> & args){
  // DATATYPE result = args[0];
  
  // for(unsigned int i=1; i<args.size(); i++){
    // result -= args[i];
  // }
  
  // result = AdjustResult(result);
  
  // return ActivateSigmoid(result);
// }

//////////////////////////////////////
// pMult class
//////////////////////////////////////
// pMult::pMult(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE pMult::evaluate(deque<DATATYPE> & args){
  // DATATYPE result = args[0];
  
  // for(unsigned int i=1; i<args.size(); i++){
    // result *= args[i];
  // }
  
  // result = AdjustResult(result);
  // return ActivateSigmoid(result);
// }

//////////////////////////////////////
// pDiv class
//////////////////////////////////////
// pDiv::pDiv(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args =number_args;
    // var_args = true;
// }

// DATATYPE pDiv::evaluate(deque<DATATYPE> & args){
  // DATATYPE result = args[0];
  
  // for(unsigned int i=1; i<args.size(); i++){
    // if(args[i] == 0){
      // result = 1.1;
      // break;
    // }
    // result /= args[i];
  // }
  
  // result = AdjustResult(result);
  
  // return ActivateSigmoid(result);
// }


//////////////////////////////////////
// pAnd class
//////////////////////////////////////

// pAnd::pAnd(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE pAnd::evaluate(deque<DATATYPE> & args){  
  // for(unsigned int i=0; i<args.size(); i++){
    // if(args[i] <= 0){
      // return 0.0;
    // }
  // }
  // return 1.0;
  
// }

//////////////////////////////////////
// pNand class
//////////////////////////////////////

// pNand::pNand(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE pNand::evaluate(deque<DATATYPE> & args){  
  // for(unsigned int i=0; i<args.size(); i++){
    // if(args[i] > 0){
      // return 0.0;
    // }
  // }
  // return 1.0;
// }

//////////////////////////////////////
// pOr class
//////////////////////////////////////

// pOr::pOr(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE pOr::evaluate(deque<DATATYPE> & args){  
  // for(unsigned int i=0; i<args.size(); i++){
    // if(args[i] > 0){
      // return 1.0;
    // }
  // }
  // return 0.0;
// }


//////////////////////////////////////
// pNor class
//////////////////////////////////////

// pNor::pNor(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE pNor::evaluate(deque<DATATYPE> & args){  
  // for(unsigned int i=0; i<args.size(); i++){
    // if(args[i] > 0){
      // return 0.0;
    // }
  // }
  // return 1.0;
// }


//////////////////////////////////////
// pXor class
//////////////////////////////////////

// pXor::pXor(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE pXor::evaluate(deque<DATATYPE> & args){ 
  // int value = (int)(args[0] > 0)?1:0;
  // int nextval;
  
  // for(unsigned int i=1; i<args.size(); i++){
    // nextval = (int)(args[i] > 0)?1:0;
    // if(nextval == value){
      // value = 1;
    // }
    // else{
      // value = 0;
    // }
  // }
  // return (DATATYPE) value;  
// }


//////////////////////////////////////
// Weight class
//////////////////////////////////////
// Weight::Weight(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE Weight::evaluate(deque<DATATYPE> & args){
    // DATATYPE result = args[0];
  
  // for(int i=1; i<num_args; i++){
    // if(args[i] == 0){
      // result = 0;
      // break;
    // } 
    // result *= args[i];
  // }
  
  // return AdjustResult(result);
// }

//////////////////////////////////////
// ConCat class
//////////////////////////////////////
// ConCat::ConCat(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// will be doubles for this implementation
// so add value of arg to 48
// '.' is 46 so that class has a value of -2
// DATATYPE ConCat::evaluate(deque<DATATYPE> & args){
  // DATATYPE result;
  
  // string newnumber;
  
  // for(unsigned int i=0; i<args.size(); i++)
    // newnumber += char(48+args[i]);

  // convert to DATATYPE
  // stringstream in(newnumber);
  // in >> result;
  
  // return result;
// }

//////////////////////////////////////
// Dot class
//////////////////////////////////////
// Dot::Dot(string symbol, int number_args)
  // :Element(symbol, true, 0){
    // num_args = number_args;
    // var_args = true;
// }

// -2 is the offset to be used with the 
// ConCat operator to generate a '.'
// DATATYPE Dot::evaluate(deque<DATATYPE> & args){
    // return -2.0;
// }

//////////////////////////////////////
// And class
//////////////////////////////////////

// And::And(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE And::evaluate(deque<DATATYPE> & args){  
  // for(unsigned int i=0; i<args.size(); i++){
    // if(args[i] <= 0){
      // return 0.0;
    // }
  // }
  // return 1.0;
  
// }

//////////////////////////////////////
// Nand class
//////////////////////////////////////

// Nand::Nand(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE Nand::evaluate(deque<DATATYPE> & args){  
  // for(unsigned int i=0; i<args.size(); i++){
    // if(args[i] > 0){
      // return 0.0;
    // }
  // }
  // return 1.0;
// }

//////////////////////////////////////
// pOr class
//////////////////////////////////////

// Or::Or(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE Or::evaluate(deque<DATATYPE> & args){  
  // for(unsigned int i=0; i<args.size(); i++){
    // if(args[i] > 0){
      // return 1.0;
    // }
  // }
  // return 0.0;
// }


//////////////////////////////////////
// pNor class
//////////////////////////////////////

// Nor::Nor(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE Nor::evaluate(deque<DATATYPE> & args){  
  // for(unsigned int i=0; i<args.size(); i++){
    // if(args[i] > 0){
      // return 0.0;
    // }
  // }
  // return 1.0;
// }


//////////////////////////////////////
// pXor class
//////////////////////////////////////

// Xor::Xor(string symbol, int number_args)
  // :Element(symbol, true, 4){
    // num_args = number_args;
    // var_args = true;
// }

// DATATYPE Xor::evaluate(deque<DATATYPE> & args){ 
  // int value = (int)(args[0] > 0)?1:0;
  // int nextval;
  
  // for(unsigned int i=1; i<args.size(); i++){
    // nextval = (int)(args[i] > 0)?1:0;
    // if(nextval == value){
      // value = 1;
    // }
    // else{
      // value = 0;
    // }
  // }
  // return (DATATYPE) value;  
// }

