// Translator.cpp

#include "Translator.h"

Translator::Translator(){
  codon_size = 8;
  word_wrap = false;
  wrap_count = 1;
	powersof2 = 0;
}


Translator::Translator(int codonsize){
  codon_size = codonsize;
  word_wrap = false;
  wrap_count = 1;
	powersof2 = 0;
	init_powersof2();
}

Translator::~Translator() {
	delete powersof2;
}

void Translator::set_grammar(Grammar* newgrammar){
  gram = newgrammar;
   // get Elements that are left and right paren and comma
  left_paren = gram->get_left_paren();
  right_paren = gram->get_right_paren();
  comma = gram->get_comma();
}

void Translator::set_word_wrap(bool wrap){
  word_wrap = wrap;
}

void Translator::set_wrap_count(int count){
  wrap_count = count;
}

void Translator::set_codon_size(int codonsize){
  codon_size = codonsize;
	init_powersof2();
	
}

void  Translator::init_powersof2() {

	//In case the codon size changes
	if (0 != powersof2) {
		delete [] powersof2;
	}

	powersof2 = new int[codon_size + 1];
	int accum = 1;
	for (int i = 0; i <= codon_size; i++) {
		powersof2[i] = accum;
		accum = 2 * accum;
	}
}

// copies codons passed in 
// to set for later evaluation
void Translator::set_codons(const vector<unsigned int> & codons){
  codon_values = codons;
}

// codon conversion adapted from libge
void Translator::convert_codons(GEBinStrBlockCrossGenome& genome){
  int num_codons = genome.length()/codon_size;
  int index =0; 
  codon_values.clear();
  
  for(int i=0; i<num_codons; i++){
    index = i*codon_size;
    codon_values.push_back(decode_binary(genome, index, codon_size));
  }  
}

vector<unsigned int> Translator::get_codon_values(){
  return codon_values;
}

// adapted from libge for calculating values of
// codons from genome
int Translator::decode_binary(GEBinStrBlockCrossGenome& genome, int start, int len){
  /*********************************************/ 
  /* Decode a binary string to numeric, from   */ 
  /* string position "start" up to start + len */ 
  /*********************************************/ 

	int accum = 0;
	int powerof2 = len-1;
  for (int j=start; j < (start+len); j++){
		accum = accum + (genome.gene(j) * powersof2[powerof2]);
    powerof2--;
  }

  return accum;
}

int Translator::create_terminals(vector<Element *> & terminal_list, 
  vector<Element *> & constants_list){
//cerr << "Translator.cpp : in create_terminals()\n";

  // clear any existing terminals from list
  terminal_list.clear();
  // tracks how many times wrapped
  int current_wrap=0, final_wrap=0;
  // track which codon being processed
  int current_codon=0, final_codon=0;
  int num_codons = codon_values.size();
  // used to create terminals list
  deque<Element *> elements_to_process;
  vector<Element *>::reverse_iterator reverse_iter;
  
  vector<Element *> elements_produced;

  // push the start nonterminal on
  elements_to_process.push_back(gram->get_start());
	
//	cerr << "Start of the grammar: " << gram->get_start() << "\n";

//	cerr << "Printing first element: " << elements_to_process[0] <<"\n";
//	cerr << "Size of elements_to_process: " << elements_to_process.size() << "\n";

  deque<DATATYPE> args;

  deque<DATATYPE> blank_args;

  vector<Element *> numElementVec, numDigVec;

  while(current_wrap < wrap_count && elements_to_process.size() > 0){
//	cerr << "Translator.cpp : Inside first while\n";
    while(elements_to_process.size() > 0 && current_codon < num_codons){
//	cerr << "Translator.cpp : Inside second while\n";
      if(elements_to_process[0]->get_isterminal()){
//	cerr << "Translator.cpp : Inside if and 2 while loops\n";
        // use the concatenation operator to create a new constant
        // that is placed on the stack
        if(elements_to_process[0] == gram->get_concat()){
          Element * concatObj = elements_to_process[0];
          elements_to_process.pop_front();
          elements_to_process.pop_front(); //skip '('
          
          numElementVec.clear();
          // convert <num> nonterminal
          if(current_codon == num_codons){
            if(++current_wrap < wrap_count){
              current_codon = 0;
            }
            else{
              break;
            }
          }
          numElementVec = gram->get_rules(codon_values[current_codon++], elements_to_process[0]);
          elements_to_process.pop_front();
          
          // now convert element list into digits
          int numArgs = int(numElementVec[numElementVec.size()-1]->evaluate());		//S: evaluate() doesn't take any arguments
          args.clear();
          for(int currDig=0; currDig < numArgs; currDig++){
            if(!numElementVec[currDig]->get_isterminal()){
              if(current_codon == num_codons){
                if(++current_wrap < wrap_count){
                  current_codon = 0;
                }
                else{
                  break;
                }
              }
              numDigVec = gram->get_rules(codon_values[current_codon++], 
                numElementVec[currDig]);
              args.push_back(numDigVec[0]->evaluate());		//S: evaluate() doesn't take any arguments
            }
            else{
              args.push_back(numElementVec[currDig]->evaluate());		//S: evaluate() doesn't take any arguments
            }
          }
          Constant * newCon = new Constant(concatObj->evaluate());		//S: evaluate() doesn't take any arguments
          terminal_list.push_back(newCon);
          constants_list.push_back(newCon);
        }
        else{
          terminal_list.push_back(elements_to_process[0]);
        }
        elements_to_process.pop_front();
      }
      else{
	//	cerr << "Translator.cpp: Inside else after 2 while loops\n";
        elements_produced = gram->get_rules(codon_values[current_codon++], elements_to_process[0]);
        elements_to_process.pop_front();
        // push in reverse order to front
        for(reverse_iter = elements_produced.rbegin(); reverse_iter != elements_produced.rend(); reverse_iter++){
          elements_to_process.push_front(*reverse_iter);
        }
      }     
    }
    final_codon = current_codon;
    final_wrap = current_wrap;
    // push terminal elements onto list to finish off elements_to_process if possible
    while(elements_to_process.size() > 0 && elements_to_process[0]->get_isterminal()){
      terminal_list.push_back(elements_to_process[0]);
      elements_to_process.pop_front();
    }

    current_codon = 0;
    current_wrap++;
  }

  if(elements_to_process.size() > 0)
	{ //cerr << "Translator.cpp : returning zero\n";
    return 0;
}
  else // return effective size of the solution
    // return 1;
    return int(final_wrap * codon_values.size() + final_codon);
}


///
/// Sets the nonterminal codon elements in the genome for
/// use in the crossover.
/// Needs to be done at start of each slave step as the first
/// crossover is done on the previous population before fitness
/// evaluation
/// No wrapping is allowed in this case
/// @param genome to set nonterminal codons for crossover
/// @return  none
///
void Translator::set_nonterminal_codons(GEBinStrBlockCrossGenome & genome){
  convert_codons(genome);
  genome.clearNonTerminalCodon();
  // track which codon being processed
  int current_codon=0;
  int num_codons = codon_values.size();
  // used to create terminals list
  deque<Element *> elements_to_process;
  vector<Element *>::reverse_iterator reverse_iter;
  vector<Element *> elements_produced;

  // push the start nonterminal on
  elements_to_process.push_back(gram->get_start());
  genome.setNonTerminalCodon(elements_to_process[0]);

  while(elements_to_process.size() > 0 && current_codon < num_codons){
    // if terminal then pop off the elements to process and 
    // look for next nonterminal
    if(elements_to_process[0]->get_isterminal()){
        elements_to_process.pop_front();
    }
    else{
      genome.setNonTerminalCodon(elements_to_process[0]);
      elements_produced = gram->get_rules(codon_values[current_codon++], elements_to_process[0]);
      elements_to_process.pop_front();
      // push in reverse order to front
      for(reverse_iter = elements_produced.rbegin(); reverse_iter != elements_produced.rend(); reverse_iter++){
        elements_to_process.push_front(*reverse_iter);
      }
    }   
  }   
}



///
/// Determines the length of a block in codons for use in
/// maintaining blocks during crossover
/// No wrapping is allowed in this case
/// @param genome
/// @param starting codon to begin checking for length
/// @param startElement starting element
/// Ret:  length of subunit or 0 if unable to complete building block
///       in available length of genome remaining
///
int Translator::determine_block_len(GEBinStrBlockCrossGenome & genome, int startCodon,
  Element * startElement){
  convert_codons(genome);

  // track which codon being processed
  int current_codon=0;
  int num_codons = codon_values.size();
  // used to create terminals list
  deque<Element *> elements_to_process;
  vector<Element *>::reverse_iterator reverse_iter;
  vector<Element *> elements_produced;

  // push the start nonterminal on
  elements_to_process.push_back(startElement);

  while(elements_to_process.size() > 0 && current_codon < num_codons){
    // if terminal then pop off the elements to process and 
    // look for next nonterminal
    if(elements_to_process[0]->get_isterminal()){
        elements_to_process.pop_front();
    }
    else{
      elements_produced = gram->get_rules(codon_values[current_codon++], elements_to_process[0]);
      elements_to_process.pop_front();
      // push in reverse order to front
      for(reverse_iter = elements_produced.rbegin(); reverse_iter != elements_produced.rend(); reverse_iter++){
        elements_to_process.push_front(*reverse_iter);
      }
    }   
  }   

  if(elements_to_process.size() > 0)
    return 0; // unable to complete building block
  else
    return current_codon;  
}

///
/// Anything except parentheses with a priority of zero is pushed to the postfix stack
/// Left paren always pushed on stack
/// When get to right paren, pop off stack until left paren encountered
/// if the symbol being scanned has a higher precedence than the symbol at the
///     top of the stack, the symbol scanned is pushed on the stack
/// If the precedence of the symbol being scanned is lower than or equal to the
///     precedence of the symbol at the top of the stack, the stack is popped to the output
/// When the terminating symbol is reached on the input scan, the stack is popped to the
///     output until the terminating symbol is also reached on the stack. Then the algorithm terminates.
/// @param input vector of Element pointers that make up a neural network
/// @return the postfix stack that can be used in the Tester to check the
/// neural network
///

//S: Not required
/*
vector<Element *> Translator::create_postfix_stack(vector<Element *> & input){
  vector<Element *> element_stack;
  
  vector<Element *> stack; // for use in creating postfix stack
  vector<Element *> postfix_stack; // stack to be returned
  int input_size = input.size();

  for(int input_count=0; input_count < input_size; input_count++){
      // priority is zero when element is a number, variable, comma or parentheses
      if(!input[input_count]->get_priority()){
        if(input[input_count] == left_paren){
          stack.push_back(input[input_count]);
        }
        else if(input[input_count] == right_paren){
          while(stack.back() != left_paren){
            postfix_stack.push_back(stack.back());
            stack.pop_back();
          }
          // pop off left paren
          stack.pop_back();
        }
        else if(input[input_count] == comma){
          while(stack.back() != left_paren){
            postfix_stack.push_back(stack.back());
            stack.pop_back();
          }
          // in this case do not pop off left paren as
          // it is being used to mark off arguments
          // to function calls
        }
        else{
          //  numbers and variables are pushed to output
          postfix_stack.push_back(input[input_count]);
        }  
      }
      else{ // element is an operator
        while(stack.size() > 0 && 
              (input[input_count]->get_priority() <= stack.back()->get_priority())){
            postfix_stack.push_back(stack.back());
            stack.pop_back();
          }
        stack.push_back(input[input_count]);
      }
  }
  
  // after last terminal push anything on the stack to the postfix stack
  for(vector<Element *>::iterator stack_iter=stack.begin(); stack_iter != stack.end(); stack_iter++)
    postfix_stack.push_back(*stack_iter);
 
  return postfix_stack;
}*/

