#!/usr/bin/perl
#
# Scott Dudek 8/5/04
# Modified to work with GP version of GPNN
#
# cross_validate.pl - Bill White - 2/23/02
#
# Perform 10-fold cross validation independent of method.
#
# All parameters are read from a config file and kept in
# a global hash. 
#
# Main functionality in separate module: CrossValidation.pm

# no output buffering
$| = 1;

use strict;
# use warnings;

use lib "/home/nicholas/motsinger/genn/lib/scripts";
use PVMController;
use CrossValidation;
#require "./PVMController.pm";
#require "./CrossValidation.pm";

my @trainCounts = ();
my @testCounts = ();

################################################## get command line parameters
if((!defined($ARGV[0])) or (scalar(@ARGV) < 1)) {
	print "\nUsage: $0 configFilename\n\n";
	exit;
}
my $configFilename = $ARGV[0];

ReportStdout("Begin Cross Validation");

############################################################# read config file
ReportStdout("*** Phase I:   Read config file");

my %configHash = ReadCVConfig($configFilename);

# get information about the dataset
GetDatasetInfo(\%configHash);

# seed the random number generator
my $randomSeed = $configHash{randomSeed};
srand($randomSeed);

# change the working directory
my $workingDirectory = $configHash{workingDirectory};
chdir $workingDirectory or
	ReportFatalError("Cannot change to working directory: " . 
									 "$workingDirectory: $!");

DumpHash(\%configHash, "CV Configuration Information");

################################################################ split dataset
ReportStdout("*** Phase II: Split dataset");
if($configHash{splitType} eq 'percentile'){
		ReportStdout(" Percentile Split");
  SplitTileDataset(\%configHash);
}
else{
		ReportStdout(" Non-percentile split");
  SplitDataset(\%configHash);
}

#################################################### write method config files
ReportStdout("*** Phase III:  Write ". 
						 $configHash{methodType} . 
						 " config files");

WriteMethodConfigFiles(\%configHash);

####################################################### train and test each CV
ReportStdout("*** Phase IV:  Training and Testing");

exit();

# if PBS_NODEFILE exists and running GPNN make PBS_HOSTFILE the hosts file 
# start PVM daemons on the VAMPIRE nodes
my $cwd = $configHash{workingDirectory};
my $pbsNodeFile = $ENV{PBS_NODEFILE};
if($configHash{methodType} eq 'GPNN' or $configHash{methodType} eq 'GP') {
        ReportStdout("Starting PVM daemons");
        my $hostsFile = $configHash{gpnnHostsFile} || $configHash{gpHostsFile};
        if(defined($pbsNodeFile)) {
                ReportStdout("Running on VAMPIRE with PBS_NODEFILE: $pbsNodeFile");
                SystemCall("cp $pbsNodeFile $hostsFile");
                # need to edit hostsfile to add .vampire to first node listed
                EditHostVamp($hostsFile);
        }
        my($error, $output) = pvmStart($hostsFile);   # start pvm
        # pvmStart() returns '1' if an error occured
        if($error) {
                ReportFatalError("Cannot start PVM daemons:\n$output");
        } else {
                ReportStdout("$output");
        }
}

my $errorFilename = 
	$configHash{baseDatasetName} . 	"." .
	$configHash{methodType} . "." . 	
	"error";
open(ERROR_FILE,">$errorFilename") or 
	ReportFatalError("Cannot open error file: $errorFilename: $!");
my ($curCV, $stdoutMsg);
my $numCrossVals = $configHash{crossValidation} || 10;
for($curCV = 1; $curCV <= $numCrossVals; $curCV++) {
	ReportStdout("***** CV - $curCV *****");
	my %resultsHash = TrainAndTestMPI(\%configHash, $curCV);
	if($resultsHash{optMisclass} == -1) {
          if($resultsHash{trnMisclass} >= 0.0001 and $resultsHash{tstMisclass} >= 0.0001){
		$stdoutMsg = sprintf("%d, %s, %6.4f, %6.4f", 
												 $curCV, 
												 $resultsHash{lociUsed}, 
												 $resultsHash{trnMisclass}, 
												 $resultsHash{tstMisclass}
												);
       }
       else{
print "USING ALTERNATE OUTPUT FROM SCRIPT\n";
         $stdoutMsg = sprintf("%d, %s, %g, %g", $curCV, $resultsHash{lociUsed},
           $resultsHash{trnMisclass}, $resultsHash{tstMisclass});
       }
	}
	else {
		$stdoutMsg = sprintf("%d, %s, %6.4f, %6.4f, %6.4f", 
												 $curCV, 
												 $resultsHash{lociUsed}, 
												 $resultsHash{trnMisclass}, 
												 $resultsHash{optMisclass},
												 $resultsHash{tstMisclass}
												);
	}
	print ERROR_FILE "$stdoutMsg\n";
	ReportStdout($stdoutMsg);
} #end for each CV run
close(ERROR_FILE);


####################################################################### report
ReportStdout("*** Phase V:   Cleaning Up Temporary Files");
Cleanup(\%configHash);

ReportStdout("End Cross Validation");

# 10/06/04
# added by smd to change name of first node listed in file, which is the node
# that will have the master on it
# a '.vampire' suffix is being added as a hostname
sub EditHostVamp{
  my $hostsFile = shift;
  open(HOST, $hostsFile) or die "Host file error: $hostsFile:  $!\n\n";
  my @hosts = <HOST>;
  chomp($hosts[0]);
  my $firsthost = $hosts[0];
  my $edited = "$hosts[0].vampire";
  close(HOST);
  open(HOST, ">$hostsFile") or die "Host file error:  $hostsFile:  $!\n\n";
  foreach my $host(@hosts){
    chomp($host);
    if($host =~ /$firsthost$/){
      print HOST "$edited\n";
    }
    else{
      print HOST "$host\n";
    }
  }
  close(HOST);
}

