#!/usr/bin/perl

#Does permutation runs of GENN in order to simulate a null distribution,
#from which a p-value can be defined.

use strict;

#Inputs:
#crossvalidate script location
#Working directory - make subdirectories for each crossvalidation run
#Template crossvalidate config file
## of permutations to run
#Original crossvalidation logfile (optional)

my ($crossvalbin, $workdir, $template, $permcount, $origlog) = @ARGV;
print "Crossvalidator: $crossvalbin\n";
print "Working directory: $workdir\n";
print "Template: $template\n";
print "Permutations: $permcount\n";
print "Original log: $origlog\n" if ('' ne $origlog);

#Seed our random generator
srand(31337);

#Only do frequency scan once
my $freqscanflag = 0;
my @statuslist = (); #original case/control status

#For each permutation:

for (my $i = 1; $i <= $permcount; $i++) {

		my $thisrs = int(rand(1000000));
		my $dataset = '';
		my $thisdata = '';
		my $datadir = '';
		my $origdata = '';

    #Create working subdir
		my $thiswd = $workdir . "/$i";
		`mkdir -p $thiswd`;

		#Copy template file into subdir and
    #modify template file:
    # workingDirectory
    # randomSeed
		# baseDatasetName
		# baseDatasetDirectory
		my $thistemp = $thiswd . "/crossvalidate.$i.config";
		open(TEMPLATEIN, "<$template") or die "Unable to open $template for reading\n";
		open(TEMPLATEOUT, ">$thistemp") or die "Unable to open $thistemp for writing\n";
		
		while (my $line = <TEMPLATEIN>) {
				chomp $line;
				if ($line =~ /^workingDirectory/) {
						$line = "workingDirectory = $thiswd";
				}
				elsif ($line =~ /^randomSeed/) {
						$line = "randomSeed = $thisrs";
				}
				elsif ($line =~ /^baseDatasetName\s*=\s*(\S+)/) {
						$dataset = $1;
						$thisdata = "perm.$i." . $dataset;
						$line = "baseDatasetName = $thisdata";
				}
				elsif ($line =~ /^baseDatasetDirectory\s*=\s*(\S+)/) {
						$datadir = $1;
						$line = "baseDatasetDirectory = $thiswd";
				}
				
				print TEMPLATEOUT "$line\n";
		}
		close(TEMPLATEIN);
		close(TEMPLATEOUT);

		$origdata = "$datadir/$dataset";
		

		#permute the last column of the base dataset
		if (!$freqscanflag) {
				open(DATAIN, "<$origdata") or die "Unable to open base dataset $origdata\n";
				
				#get case/control frequencies (last column)
				my $casecount = 0;
				my $controlcount = 0;
				while (my $line = <DATAIN>) {
						
						chomp $line;
						my ($char) = ($line =~ /.*\s+(\d+)/);
						if ('1' eq $char) {
								$casecount++;
						}
						elsif ('0' eq $char) {
								$controlcount++;
						}
						else {
								die "Unrecognized case/control status: \"$char\"\n";
						}
						push(@statuslist, $char);
				}
				
				my $controlfreq = 1.0 * $controlcount / ($casecount + $controlcount);
				print "Control frequency: " . $controlfreq . "\n";
				$freqscanflag = 1;
				close(DATAIN);
		}
		
		#randomly generate with same frequencies and output to new dataset

		my @thesestati = @statuslist;
		shuffle(\@thesestati);

		open(DATAIN, "<$origdata") or die "Unable to open base dataset $origdata\n";
		open(DATAOUT, ">$thiswd/$thisdata") or die "Unable to open new dataset $thiswd/$thisdata\n";
		my $counter = 0;

		#print "thisdata $thiswd/$thisdata\n";
		
		while (my $line = <DATAIN>) {
				chomp $line;
				
				if ($thesestati[$counter] eq '') {
						die "blank status at index $counter\n";
				}
				
				$line =~ s/\d+$/$thesestati[$counter]/;

				$counter++;

				print DATAOUT "$line\n";

		}
		close(DATAIN);
		close(DATAOUT);

		#run cross_validate, capture output in a logfile
		print "Doing permutation $i...\n";
		my $cmd = "$crossvalbin $thistemp > $thiswd/xval.log";
		#print "$cmd\n";
		`$cmd`;

}

#After all permutations are done:
#crunch each logfile to find testing misclassification rate

print "Determining distribution...\n";

my @avgdist = ();
for (my $i = 1; $i <= $permcount; $i++) {
		my $thiswd = $workdir . "/$i";
		my $thislog = "$thiswd/xval.log";

		push(@avgdist, getError($thislog));

}

#Output sorted distribution
@avgdist = sort { $a <=> $b } @avgdist;
open(DISTOUT, ">$workdir/distribution.out") or 
		die "Unable to open $workdir/distribution.out\n";

foreach my $avg (@avgdist) {
		print DISTOUT "$avg\n";
}
close(DISTOUT);


#Locate original rate in distribution, output p-value (if original given).





#Parse a xval log file to get misclassification error
#Single argument a filename, return value of average
# testing misclassification
sub getError {
		my $filename = $_[0];
		my @lines = `grep 'Testing misclassification' $filename`;
		my $sum = 0;
		my $count = 0;
		foreach my $line (@lines) {
				chomp $line;
				my ($error) = ($line =~ /Testing misclassification: (\S+)/);
				$sum += $error;
				$count++;
		}
		my $avg = 1.0 * $sum / $count;
		return $avg;
}




#randomize an array - code pulled from somewhere online
sub shuffle {
    my $array = shift;
    my $i = scalar(@$array);
    my $j;
    foreach my $item (@$array )
    {
        --$i;
        $j = int rand ($i+1);
        next if $i == $j;
        @$array [$i,$j] = @$array[$j,$i];
    }
    return @$array;
}
