#!/usr/bin/perl

#Takes an input file with the first column class, and succeeding columns being
#genotype data (AA/AT/TA/etc).  Transforms these into numeric values (0/1/2/3/4)
#for each column.

use strict;

my $inputfile = 'mid_coded_sub_orig_selected.TXT';
my $outputfile = "$inputfile.recoded";

open(INFILE, "<$inputfile") or die "Unable to open $inputfile for reading\n";
open(OUTFILE, ">$outputfile") or die "Unable to open $outputfile for writing\n";

my %data = (); #$data{individual}{locus}
my %data2 = (); #$data{individual}{locus} (numberized)
my $counter = 0; #number of individuals
my $colcount = -1; #

#read in the data
while (my $line = <INFILE>) {
		chomp $line;
		$line =~ s/[\000-\010]//g;
		$line =~ s/[\012-\037]//g;
		#print "$line\n";
		my @tmp = split("\t", $line);
		$colcount = @tmp;
		for (my $i = 0; $i < $colcount; $i++) {
				$data{$counter}{$i} = $tmp[$i];
		}
		$counter++;
} #end while infile
close(INFILE);

print "colcount: $colcount\n";
print "counter: $counter\n";


#convert genotypes to indicators on a per-column basis
#this could be done in a single hash, but isn't.
for (my $i = 1; $i < $colcount; $i++) {
		my %genohash = (); #genotypes we've seen before
		my $genonum = 0;  #genotype number to assign next
		$genohash{'.'} = '.'; #override for missing data
		for (my $j = 0; $j < $counter; $j++) {
				
				#print "Looking at " . $data{$j}{$i} . "\n";
				#if it's new, give it a number.  For association purposes, genotype
				# CT is the same as TC.
				if (!($data{$j}{$i} =~ /^[ACGT]{2}$/)) {
						print STDERR "Warning - odd looking genotype " . $data{$j}{$i} .
								" for column $i, row $j\n";
				}
				if ($genohash{$data{$j}{$i}} eq '' || $genohash{reverse($data{$j}{$i})} eq '' ) {
						#this is a new one
						print "Dubbing " . $data{$j}{$i} . " to be $genonum\n";
						$genohash{$data{$j}{$i}} = $genonum;
						$genohash{reverse($data{$j}{$i})} = $genonum;
						$genonum++;
				}

				$data2{$j}{$i} = $genohash{$data{$j}{$i}};
		}
}


#output converted data
for (my $j = 0; $j < $counter; $j++) {
		for (my $i = 0; $i < $colcount; $i++) {
				if ($i == 0) {
						print OUTFILE $data{$j}{$i};
				}
				else {
						print OUTFILE $data2{$j}{$i};
				}
				print OUTFILE "\t" if ($i < $colcount - 1);
		}
		print OUTFILE "\n";
}
close(OUTFILE);
