// $Header: /usr5/cvsroot/nprojects/motsinger/genn/genn/src-mpi/GARandomSearchGA.h,v 1.2 2007/10/17 14:56:22 cvs Exp $
/* ----------------------------------------------------------------------------
  GARandomSearch.h
  mbwall 28jul94
  Copyright (c) 1995 Massachusetts Institute of Technology
                     all rights reserved

  Header file for the simple genetic algorithm class.
---------------------------------------------------------------------------- */

///
/// Modified to do random GENN search
///

#ifndef _ga_garandomsearch_h_
#define _ga_garandomsearch_h_

#include <ga/GABaseGA.h>
#include <vector>
#include <iostream>
#include "genome.h"

class GARandomSearchGA : public GAGeneticAlgorithm {
public:
  GADefineIdentity("GARandomSearchGA", GAID::SimpleGA);

  static GAParameterList& registerDefaultParameters(GAParameterList&);

public:
  GARandomSearchGA(const GAGenome&);
  GARandomSearchGA(const GAPopulation&);
  GARandomSearchGA(const GARandomSearchGA&);
  GARandomSearchGA& operator=(const GARandomSearchGA&);
  virtual ~GARandomSearchGA();
  virtual void copy(const GAGeneticAlgorithm&);

  virtual void initialize(unsigned int seed=0);
  virtual void step();
  GARandomSearchGA & operator++() { step(); return *this; }

  virtual int setptr(const char* name, const void* value);
  virtual int get(const char* name, void* value) const;

  virtual int minimaxi() const {return minmax;}
  virtual int minimaxi(int m);

  virtual const GAPopulation& population() const {return *pop;}
  virtual const GAPopulation& population(const GAPopulation&);
  virtual int populationSize() const {return pop->size();}
  virtual int populationSize(unsigned int n);
  virtual GAScalingScheme& scaling() const {return pop->scaling();}
  virtual GAScalingScheme& scaling(const GAScalingScheme & s)
    {oldPop->scaling(s); return GAGeneticAlgorithm::scaling(s);}
  virtual GASelectionScheme& selector() const {return pop->selector(); }
  virtual GASelectionScheme& selector(const GASelectionScheme& s)
    {oldPop->selector(s); return GAGeneticAlgorithm::selector(s);}
  virtual void objectiveFunction(GAGenome::Evaluator f);
  virtual void objectiveData(const GAEvalData& v);

  bool debugOn;

  void debugStreamSet(string outfilename){debugstream.open(outfilename.c_str(), ios::out);}

protected:
  GAPopulation *oldPop;		// current and old populations
  GABoolean el;			// are we elitist?
  int effectivePopSize; // holds effective population size for random search
  
  float fitnessEvaluator(std::vector<unsigned int> & codons);
  float randomGenomeCreate(std::vector<unsigned int> & codons, bool fullTree);
  
  ofstream debugstream;
  void outputIndividual(vector<unsigned int> & codons, int currInd, float fitness);
  void outputIndividual(GAGenome & genome, int currInd);
};



#ifndef NO_STREAMS
inline ostream& operator<< (ostream& os, GARandomSearchGA & arg)
{ arg.write(os); return(os); }
inline istream& operator>> (istream& is, GARandomSearchGA & arg)
{ arg.read(is); return(is); }
#endif

#endif
