//Grammar.h

#ifndef __GRAMMAR_H__
#define __GRAMMAR_H__

#include "Terminals.h"
#include "ErrorExcept.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <vector>

/// contains rule information
struct Rule_status{
  bool isrecursive;
  int  minSteps;
};

using namespace std;

///
/// Parses a grammar file and creates the grammar rules
/// using the Element class and sub-classes
///

class Grammar{
  
  public:
    /// creates instances of syntactical elements such as commas
    Grammar();
    ~Grammar();
    
    /// Parses grammar file and creates rules for production of individuals
    void read_gram_file(string filename);  
    
    // Checks that string is a number
    static bool isnumber(string str);
    
    Element * get_start() const;
    Element * get_left_paren();
    Element * get_right_paren();
    Element * get_comma();
    Element * get_concat(){ return concatOp;}
   
    /// returns vector that contains elements for rule
    vector<Element *> get_rules(int codon_value, Element * el);
    
    /// calculates values for use in sensible initialization
    void calculate_rule_values(int maxDepth); 
    
    int get_num_recursive_opts(Element* ruleElement, int depthLeft);
    int get_num_opts(Element* ruleElement, int depthLeft);
    int get_real_recursive_index(Element* ruleElement, int depthLeft, int randIndex);
    int get_real_index(Element* ruleElement, int depthLeft, int randIndex);
    int get_num_rules(Element* ruleElement);
    int get_init_tree_depth() const;
    
  private:
    void insert_spaces(string & line);
    int split_string(const string& input, 
      string delim1, string delim2, vector<string>& results);
    void set_nonterminals(vector<string> & tokens);
    void set_start(vector<string> & tokens);
    void create_rules(vector<string> & lines);
    void set_terminals(vector<string> & tokens);
   
    bool check_recursion(Element* startElement, 
        map<Element *, bool> searchElements);
        
    int calc_element_min_depth(Element* currentElement, 
      map<Element*, bool> prevElements, int argDepth);
      
    Terminal_creator terminal_machine;
    map<string, Element* > element_map; // used in creating rules 
    map< Element*, vector<vector<Element*> > > rules; // production rules
    
    map< Element*, vector<Rule_status> > rule_values; // used for sensible initialization
    map< Element*, bool> recursiveNonTerms; // used to track which nonterminals are recursive
    map< Element*, vector<vector<int> > >  valid_rule_indices_all;
    map< Element*, vector<vector<int> > > valid_rule_indices_recursive;
    
    int initTreeDepth;
    Element* start_codon;
    Element* concatOp;
};


#endif   
