//Terminals.h

// This file should contain the headers for all terminals
// All the terminals must inherit from the Element class

#ifndef __TERMINALS_H__
#define __TERMINALS_H__

#include "Element.h"
#include <map>
#include "ErrorExcept.h"
#include <sstream>
#include <stdlib.h>
#include "Stringmanip.h"
#include "Config.h"
using namespace std;


/// Terminal holder creates and returns the Elements by name
class Terminal_creator{
  public:
    Terminal_creator();
    Element * create_terminal(const string& symbol, int num_args=0);
    Element * create_constant(const string& symbol);
    
  private:
    /// used for variable creation (variables defined as V1, V2, V3, V4
    Element * create_variable(const string& symbol, int var_index);
    
    /// add to enumeration below for each additional terminal
    enum TerminalList{ 
        NotDefined,
        Add,
        Sub,
        Mult,
        Div,
        padd,
        psub,
        pmult,
        pdiv,
        wt,
        dot,
        var,
        concat,
        pand,
        pnand,
        por,
        pnor,
        pxor,
        booland,
        boolnand,
        boolor,
        boolnor,
        boolxor
    };
    
    map<string, TerminalList> TerminalMap;
};


class Constant :public Element{
  
  public:
    Constant(string symbol, int number_args=0);
    Constant(DATATYPE val);
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);

    private:
      DATATYPE constant_value;
};

///////////////////////////////
// user-defined terminals below
///////////////////////////////

///
/// Element that contains an index for referencing 
/// genotypes in the dataset
///
class Variable :public Element{
  
  public:
    Variable(string symbol, int number_args=0);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
    static void fill_variable_list(vector<INPUT_TYPE>  variables);
		static void fill_variable_list_bits(vector<BITTYPE> variablebits);
    static void set_ind_index(int new_index);
    static void clear_variable_list();
    static void set_missing_value(int miss_value);
		static void set_memory_manager(MemoryManagers mmanager);
 private:
		int index_value;
		static vector< vector <INPUT_TYPE> > var_values;
		static vector< vector <BITTYPE> > var_values_bits;
		static int ind_index, missing_value;

		/*Store which type of memory manager we're using, so as to
			properly access variables*/
		static MemoryManagers memory_manager;
		static int varspernum; //number of INPUT_TYPE per BITTYPE
		
};

class Addition :public Element{
  
  public:
    Addition(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class Subtraction :public Element{
  
  public:
    Subtraction(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class Multiplication :public Element{
  
  public:
    Multiplication(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class Division :public Element{
  
  public:
    Division(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class pAdd :public Element{
  
  public:
    pAdd(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class pSub :public Element{
  
  public:
    pSub(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class pMult :public Element{
  
  public:
    pMult(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class pDiv :public Element{
  
  public:
    pDiv(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class Weight :public Element{
  
  public:
    Weight(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};


class Dot :public Element{
  
  public:
    Dot(string symbol, int number_args=0);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class ConCat :public Element{

  public:
    ConCat(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
    friend ostream & operator << (ostream & os, const Element & el);
};
   
// Following are the boolean operators for neural nets
class pAnd :public Element{
  
  public:
    pAnd(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class pNand :public Element{
  
  public:
    pNand(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class pOr :public Element{
  
  public:
    pOr(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class pNor :public Element{
  
  public:
    pNor(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class pXor :public Element{
  
  public:
    pXor(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

// Following are the boolean operators
class And :public Element{
  public:
    And(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class Nand :public Element{
  
  public:
    Nand(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class Or :public Element{
  
  public:
    Or(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class Nor :public Element{
  
  public:
    Nor(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};

class Xor :public Element{
  
  public:
    Xor(string symbol, int number_args);
    
    virtual DATATYPE evaluate(deque<DATATYPE> & elements);
};


#endif
