//Tester.cpp

#include "Tester.h"

///
/// In the beginning, there were none
///
int Tester::missing_count = 0;

///
/// Add an individual status value 
///
void Tester::add_status(int stat){
  ind_status.push_back(stat);
}

///
/// Add an individual trait value
///
void Tester::add_status_quan(DATATYPE trait) {
	ind_status_quan.push_back(trait);
}

///
/// Clear all the individuals' statuses in the Tester
///
void Tester::clear_status(){
  ind_status.clear();
	ind_status_quan.clear();
}


///
///  Returns the error rate of a particular network that
///  is represented as a set of operators in postfix notation
///  @param postfix vector of operators
///  @return error rate
///
double Tester::determine_error_rate(vector<Element *> & postfix){

  int vector_size = ind_status.size();
  int postfix_size = postfix.size();
  vector<DATATYPE> stack;
  
  unsigned int caseright=0, controlright=0;
  
  deque<DATATYPE> args;
  int num_args=0;
	int total_right=0;
  int result;
	vector <DATATYPE> results_quan;
	vector <DATATYPE> data_quan;

  
  // for each individual evaluate the model
  // based on the postfix stack
  for(int i=0; i<vector_size; i++) {
    Variable::set_ind_index(i);
    try{
			for(int j=0; j<postfix_size; j++){
				args.clear();
				//evaluate and push any elements that don't take arguments
				if(!(num_args=postfix[j]->get_num_args())){
					stack.push_back(postfix[j]->evaluate(args));
				}
				else{
					// need to determine number of arguments
					// pop off arguments and then push back new result
					// when number of args for the element is < 0 
					// top number on stack should be the number of 
					// arguments for the function
					if(num_args<0){
						num_args = int(stack.back());
						stack.pop_back();
					}
					// construct vector of arguments
					for(int k=0; k<num_args; k++){
						args.push_front(stack.back());
						stack.pop_back();
					}
					
					// evaluate current operator and push onto stack
					stack.push_back(postfix[j]->evaluate(args));
				}
			}
			result = ((stack.back() > 0.5)?1:0);

			results_quan.push_back(stack.back());
			data_quan.push_back(ind_status_quan[i]);
			
			// add balanced accuracy calculation 
			// 7/14/06
			//nihard - fix it on 10/24/08
			if(result == ind_status[i]) {
				//total_wrong++;
				total_right++;
				// if affected increment casewrong
				if(ind_status[i]) {
					caseright++;
				}
				else {
					controlright++;
				}
			}
		} 
		catch(ErrorExcept & e){
			if(e.code() != ErrorExcept::missingData)
				throw(e);
			else{
				if (missing_count < MAX_MISSING) {
					cout << "missing found and ignored" << endl;
					missing_count++;
				}
			}
		}
		
  } /*End eval for each individual*/ 
	
	    
	/*Determine total number of cases/controls*/
	int num_cases = 0, num_controls = 0;
	for (int i = 0; i < vector_size; i++) {
		if (ind_status[i]) {
			num_cases++;
		}
		else {
			num_controls++;
		}
	} /*End foreach individual*/
	
	int casewrong, controlwrong;
	
	casewrong = num_cases - caseright;
	controlwrong = num_controls - controlright;
	
	int total_wrong;
	total_wrong = casewrong + controlwrong;

  // sensitivity = a/(a+c)
  // specificity = d/(b+d)
  if(fitnessType == classError){  
    return double(total_wrong)/vector_size;
  }
  // balanced accuracy is the average of sensitivity and specificity
  else if (fitnessType == balancedAcc) {
    return 1-.5 * (double(caseright) / (caseright + casewrong) + 
    double(controlright) /(controlright + controlwrong));
  }
	else if (fitnessType == weightedCE) {
		
		double retval =  (((double(vector_size) / num_cases) * casewrong +
													(double(vector_size) / num_controls) * controlwrong) /
												 (2 * vector_size));
		
		/*cerr << "num cases: " << num_cases << " num controls: " << num_controls <<
			" casewrong: " << casewrong << " controlwrong: " << controlwrong << 
			" weightedCE: " << retval << "\n"*/;
		//double classE = double(total_wrong)/vector_size;
		//cerr << "classE: " << classE << " weightedCE: " << retval << "\n";
		return retval;

	} /*End weightedCE*/
	else if (fitnessType == rSquared) {

		/*We don't check the underlying data, just the trait values that were
			stored with the predicted ones.  This way, we skip NNs that didn't
			evaluate due to missing values*/

		unsigned int known_size = data_quan.size();

		
		/*From "Numerical Recipes in C++"*/
		const double TINY = 1.0e-20;  //regularize complete correlation
		double yt, xt, r;
		double syy=0.0, sxy=0.0, sxx=0.0, ay=0.0, ax=0.0;
		int j;
		
		//Find means
		for (j = 0; j < known_size; j++) {
			ax += data_quan[j];
			ay += results_quan[j];
		}
		ax /= known_size;
		ay /= known_size;
		
		//Get R
		for (j = 0; j < known_size; j++) {
			xt = data_quan[j] - ax;
			yt = results_quan[j] - ay;
			sxx += xt * xt;
			syy += yt * yt;
			sxy += xt * yt;
		}
		r = sxy / (sqrt(sxx * syy) + TINY);
		
		//check for impossible r
		if ( (r < -1.0) || (r > 1.0) ) {
			cerr << " Something odd:\n";
			cerr << "r: " << r << " ax: " << ax << " ay: " << ay << " sxx: " << 
				sxx << " syy: " << syy << " sxy: " << sxy << "\n";
			
			for (int i = 0; i < vector_size; i++) {
				cerr << "status: " << data_quan[i] << " pred: " << results_quan[i] << "\n";
			}
		} //end if impossible r



		return (double)(1 - (r*r));


	}
	else {
		cerr << "ERROR:  unknown fitness type\n";
		exit(1);
	}
}


///
/// Sets fitness type so either classification error or
/// balanced accuracy will be returned
/// @param fitType enum lising fitness type to be used
///
void Tester::set_fitness_type(FitnessTypes fitType){
  fitnessType = fitType;
}
