//Translator.h

#ifndef __TRANSLATOR_H__
#define __TRANSLATOR_H__

#include "Grammar.h"
#include "ErrorExcept.h"
#include <iostream>
#include <vector>
#include <deque>
#include <ga/ga.h>
#include "GEBinStrBlockCrossGenome.h"

using namespace std;

///
/// Translates the binary string into
/// a phenotype of codons
///

class Translator{
  
  public:
    /// creates instances of syntactical elements such as commas
    Translator(int codonsize); 
    Translator();

		~Translator();
    
    /// sets reference to grammar to use in translating genome
    void set_grammar(Grammar * newgrammar);
    
    /// converts binary genome into integer vector stored in codon_values
    void convert_codons(GEBinStrBlockCrossGenome & genome);
    
    /// sets codons within this object for later evaluation
    void set_codons(const vector<unsigned int> & codons);
    
    /// uses grammar and chromosome to create a list of terminals
    /// terminal_list is used to store the list
    /// return 0 if not a valid individual created
    int create_terminals(vector<Element *> & terminal_list, vector<Element *> & constants_list);
    
    /// converts terminal_list into postfix stack for use 
    /// in evaluating results
    vector<Element *> create_postfix_stack(vector<Element *> & terminal_list);
    
    void set_word_wrap(bool wrap);
    void set_wrap_count(int count);
    void set_codon_size(int codonsize);
    vector<unsigned int> get_codon_values();
   
    int determine_block_len(GEBinStrBlockCrossGenome & genome, int startCodon, Element * startElement);
    
    void set_nonterminal_codons(GEBinStrBlockCrossGenome & genome);
   
  private:
 
    int decode_binary(GEBinStrBlockCrossGenome& genome, int start, int len);
		void init_powersof2();
  
    Grammar* gram;
    vector<unsigned int> codon_values;
    int codon_size, wrap_count;
    bool word_wrap;
    Element * left_paren, * comma, * right_paren;

		//Lookup tables ftw
		int * powersof2;

};


#endif   

